'' MCP3208 Fast Multiple Chip Automatic Sample Driver

'' Modification by SRLM, August 2009

'' MCP3208_fast_multi_autosample.spin
''
'' Copyright (c) 2008, 2009 Jim, (c) 2009 SRLM
''
'' See end of file for terms of use.
''
''*****************************************************
''*  MCP3208 12-bit/8-channel ADC Driver              *
''*  Works with up to three MCP3208 chips             *
''*  Provides single sample or average of n samples   *
''*  Provides single mode only, no differential mode  *
''*  Does not provide DAC output like original code   *
''*  Automatically samples all 24 channels            *
''*  Requires 25 longs from the user (dataaddr points)*
''*  2KHz sample rate for single average at 80MHz clk *
''*  To change num samples see "numsamples" in DAT blk* 
''*****************************************************
                                    
'' rev 1.1   12-26-08  fixed comments
'' rev 1.2   01-15-09  fixed bugs in div( ) and div_ routines
''- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
'' rev 1.3   08-11-09  added (by SRLM) autosampling with iteration count

VAR
  long  cog

OBJ
 ' dbg   :       "PASDebug"                '<---- Add for Debugger

PUB start(dataaddr, dpin1, cpin1, spin1, dpin2, cpin2, spin2, dpin3, cpin3, spin3) : okay

'' Start driver - starts a cog
'' returns false if no cog available
'' may be called again to change settings (and erase cycles count)
''
''
''   dataaddr = pointer to a cycles variable (long) immediately followed by 24 long array
''
''   dpin  = pin connected to both DIN and DOUT on MCP3208
''   cpin  = pin connected to CLK on MCP3208
''   spin  = pin connected to CS  on MCP3208

''   dpin1 = dpin for 1st MCP3208, etc.
  
  stop                          ' stop cog if running from before

  longmove(dataaddr, @dpin1, 9)

  'dbg.start(31,30,@entry)                 '<---- Add for Debugger 

  'cog := cognew(@entry, dataaddr) + 1  ' store adc_read and shared mem address
  cog := cognew(@setup_, dataaddr) + 1  ' store adc_read and shared mem address 
  waitcnt(clkfreq/2 + cnt)      ' wait 1/2 sec for cog to start
  
  return cog
   

PUB stop

'' Stop driver - frees a cog

  if cog
    cogstop(cog~ - 1)

    

  
DAT
''
''************************************
''* Assembly language MCP3208 driver *
''************************************


                        org     0
       
{   
entry

'  --------- Debugger Kernel add this at Entry (Addr 0) ---------
   long $34FC1202,$6CE81201,$83C120B,$8BC0E0A,$E87C0E03,$8BC0E0A
   long $EC7C0E05,$A0BC1207,$5C7C0003,$5C7C0003,$7FFC,$7FF8
'  --------------------------------------------------------------        }

'************************************
setup_
                        mov     resultptr, #values      'Store the address of where to store the results
                        
                        ' Get arguments (pin definitions)
                           
                        mov     t3, par
                        rdlong  t1, t3
                        call    #param                  ' setup DIN/DOUT1 pin
                        mov     dmask1,t2

                        add     t3, #4
                        rdlong  t1, t3
                        call    #param                  ' setup CLK1 pin
                        mov     cmask1,t2

                        add     t3, #4
                        rdlong  t1, t3
                        call    #param                  ' setup CS1 pin
                        mov     smask1,t2

                        add     t3, #4
                        rdlong  t1, t3
                        call    #param                  ' setup DIN/DOUT2 pin
                        mov     dmask2,t2

                        add     t3, #4
                        rdlong  t1, t3     
                        call    #param                  ' setup CLK2 pin
                        mov     cmask2,t2

                        add     t3, #4
                        rdlong  t1, t3
                        call    #param                  ' setup CS2 pin
                        mov     smask2,t2

                        add     t3, #4
                        rdlong  t1, t3
                        call    #param                  ' setup DIN/DOUT3 pin
                        mov     dmask3,t2

                        add     t3, #4
                        rdlong  t1, t3
                        call    #param                  ' setup CLK3 pin
                        mov     cmask3,t2

                        add     t3, #4
                        rdlong  t1, t3
                        call    #param                  ' setup CS3 pin
                        mov     smask3,t2

                        'The follow code sets up delayval, from delay_ routien in original code
                        'The 40 is calculated for an 80MHz clock or less
                        mov     t4, #40                 ' set up delay time for chip clock
                        shr     t4, #3                  ' divide it by 8 clock cycles/ 2 instr
                        sub     t4, #4                  ' adj for 1/2 cycle
                        mov     delayval, t4            ' store in delayval for delay routines
                        
                        
                        jmp     #average_

'************************************
param                   mov     t2,#1                   ' make pin mask in t2
                        shl     t2,t1
param_ret               ret

'************************************
average_                mov     t3, numsamples          ' set for n sample average (or 1 if single sample)
                         
                        cmp     chip, #2        wc,wz   ' test to see if arg1 (chipnum) is 1, 2 or 3

              if_z      mov     dmask, dmask2           ' set for chip #2 if z bit set
              if_z      mov     cmask, cmask2
              if_z      mov     smask, smask2
              if_z      jmp     #:calc_adc

              if_c      mov     dmask, dmask1           ' set for chip #1 if carry set
              if_c      mov     cmask, cmask1
              if_c      mov     smask, smask1
              if_c      jmp     #:calc_adc

                        mov     dmask, dmask3           ' set for chip #3 if neither bit set
                        mov     cmask, cmask3
                        mov     smask, smask3

:calc_adc               or      dira,cmask              ' set CLK line to output
                        or      dira,smask              ' set CS  line to output

                        mov     command,#$18            ' init command to 011000, start/single
                        add     command, channel        ' add in channel number to command bits
                        
                        mov     t2, #0                  ' initialize data[channel] accumulator
                        
:bloop                  mov     stream,command          ' set up for new sample
                        or      outa,smask              ' CS high
                        or      dira,dmask              ' make DIN/DOUT output
                        andn    outa,cmask              ' CLK MUST be low before CS enabled!!!
                        andn    outa,smask              ' CS low
                        nop                             ' wait 8 clock cycles for TSUCS              
                        nop
                        
                        mov     bits,#5                 ' ready 20 bits (cs+1+diff+ch[3]+0+0+data[12])
:cloop                  test    stream,#$10     wc      ' update DIN/DOUT
                        muxc    outa,dmask
                        andn    outa,cmask              ' CLK low
                        call    #delay1                 ' adjust clock cycle time                        
                        rcl     stream,#1               ' shift out next bit
                        or      outa,cmask              ' CLK high                        
                        call    #delay2                 ' adjust clock cycle time                        
                        djnz    bits,#:cloop

                        andn    dira,dmask              ' make DIN/DOUT input to get data
                        mov     bits,#14                ' get the data 2-null plus 12 data
:dloop                  andn    outa,cmask              ' CLK low                        
                        call    #delay1                 ' adjust clock cycle time                        
                        or      outa,cmask              ' CLK high
                        test    dmask,ina       wc      ' sample DIN/DOUT
                        rcl     stream,#1               ' store bit for output         
                        call    #delay2                 ' adjust clock cycle time                        
                        djnz    bits, #:dloop           ' next data bit

                        and     stream,mask12           ' trim sample
                        add     t2, stream              ' accumulate sum in t2
                        djnz    t3, #:bloop             ' more samples for average?

                        cmp     numsamples, #1        wz' see if n > 1
        if_nz           mov     t1, numsamples          ' set up for divide if needed, arg2 = number of samples, n
        if_nz           call    #divide                 ' divide t2 by t1 or sum/n -> t2

                        
                                               
                       ' add     debugnumber, #1
                       ' mov     t2, debugnumber

                        movd    :result_loc, resultptr  ' modify the next destination
                        nop                             ' nop to allow for self modifying code
:result_loc             mov     0-0, t2                 ' store the result into the values array 
                        add     resultptr, #1           ' increment result pointer
                        add     channel, #1             ' increment channel number
                        cmp     channel, #8        wz   ' test for channel 8 (not possible)
        if_z            add     chip, #1                ' roll over channel to 0, chip++
        if_z            sub     channel, #8
                        cmp     chip, #4          wz    ' test for chip 4 (not possible)
        if_z            sub     chip, #3                ' roll over chip to 1
        if_z            mov     resultptr, #values      ' reset result pointer to values[0]
        if_z            jmp     #write_results          ' if the chip has rolled over, then write results to hub
                        jmp     #average_

        '************************************
write_results
                        't1  = hub pointer
                        '0-0 = cog pointer
                        't3  = loop counter

                        mov     t1, par                 ' Get base address in Hub
                        add     t1, #4                  ' Move to start of array in Hub
                        movd    :wr_loop, #values             ' Get Cog array address and store in instruction
                        mov     t3, #24
                        
:wr_loop                wrlong  0-0, t1                  ' 1 Write the value to the hub
                        add     t1, #4
                        add     :wr_loop, wrloopmask                        
                        djnz    t3, #:wr_loop

                        
                        add     count, #1
                        wrlong  count, par
                        
                        jmp     #average_               ' Return to take more samples
                        
                        
'************************************
delay1                  mov     t4, delayval            ' load delayval, clock cycles
                        nop                             ' adj to half cycle
                        nop
:tloop1                 nop
                        djnz    t4, #:tloop1            ' delay 1/2 uSec = 1/2 MCP3208 clk cycle
delay1_ret              ret
                        
'************************************
delay2                  mov     t4, delayval            ' load delayval, clock cycles
:tloop2                 nop
                        djnz    t4, #:tloop2            ' delay 1/2 uSec = 1/2 MCP3208 clk cycle
delay2_ret              ret
                        
'************************************
' Divide                
'   in:         t1 = 16-bit divisor  in long
'               t2 = 16-bit dividend in long
                        
'   out:        t2 = 16-bit quotient (in long), truncated
                        
'   temp:       t3 = loop counter
                        
divide                  shl     t1,#15                  ' get divisor into t1[30..15]
                        mov     t3,#16                  ' loop counter for 16 bits
                        
:loop                   cmpsub  t2,t1 wc                ' if t1 =< t2 subtract it, quotient bit -> c bit
                        rcl     t2,#1                   ' rotate c into quotient and shift dividend
                        djnz    t3,#:loop               ' loop until done
                        
                        and     t2, hexFFFF             ' mask quotient, drop the remainder in [31..16]
divide_ret              ret                             

'************************************
' Initialized data
'
hex200                  long    $200                    ' constants
hexFFFF                 long    $FFFF
mask12                  long    $FFF
zero                    long    0

numsamples              long    14                      ' The number of samples to average, up to 16(USER EDITABLE)
chip                    long    1                       ' The working register to dictate which chip
channel                 long    0                       ' The working register to dictate which channel
count                   long    0                       ' The current count of writes to the hub
wrloopmask              long    $200                    ' Will add 1 to a cog address in the destination field

debugnumber             long    1


'************************************
' Uninitialized data
'
dmask                   res     1                       ' working mask registers
cmask                   res     1
smask                   res     1

dmask1                  res     1                       ' MCP3208 pin masks for chip #1
cmask1                  res     1
smask1                  res     1

dmask2                  res     1                       ' MCP3208 pin masks for chip #2
cmask2                  res     1
smask2                  res     1

dmask3                  res     1                       ' MCP3208 pin masks for chip #3
cmask3                  res     1
smask3                  res     1

t1                      res     1                       ' temps
t2                      res     1
t3                      res     1
t4                      res     1

command                 res     1                       ' for MCP3208 routine
stream                  res     1
bits                    res     1
delayval                res     1


resultptr               res     1                       ' Where to store the result of an average read (answer: in the value array)
values                  res     24                      ' The array of results to write to the hub

                        FIT     495
{ Terms of use:

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}